@@ -14,44 +14,53 @@ import copy
 import os
 import shutil
 import tempfile
+from io import BytesIO
 
-from tests import BaseTaskTest
-from tests import BaseSubmissionTaskTest
-from tests import StreamWithError
-from tests import FileCreator
-from tests import unittest
-from tests import RecordingExecutor
-from tests import NonSeekableWriter
-from s3transfer.compat import six
+from s3transfer.bandwidth import BandwidthLimiter
 from s3transfer.compat import SOCKET_ERROR
+from s3transfer.download import (
+    CompleteDownloadNOOPTask,
+    DeferQueue,
+    DownloadChunkIterator,
+    DownloadFilenameOutputManager,
+    DownloadNonSeekableOutputManager,
+    DownloadSeekableOutputManager,
+    DownloadSpecialFilenameOutputManager,
+    DownloadSubmissionTask,
+    GetObjectTask,
+    ImmediatelyWriteIOGetObjectTask,
+    IOCloseTask,
+    IORenameFileTask,
+    IOStreamingWriteTask,
+    IOWriteTask,
+)
 from s3transfer.exceptions import RetriesExceededError
-from s3transfer.download import DownloadFilenameOutputManager
-from s3transfer.download import DownloadSeekableOutputManager
-from s3transfer.download import DownloadNonSeekableOutputManager
-from s3transfer.download import DownloadSubmissionTask
-from s3transfer.download import GetObjectTask
-from s3transfer.download import IOWriteTask
-from s3transfer.download import IOStreamingWriteTask
-from s3transfer.download import IORenameFileTask
-from s3transfer.download import CompleteDownloadNOOPTask
-from s3transfer.download import DeferQueue
-from s3transfer.futures import IN_MEMORY_DOWNLOAD_TAG
-from s3transfer.futures import BoundedExecutor
-from s3transfer.utils import OSUtils
-from s3transfer.utils import CallArgs
+from s3transfer.futures import IN_MEMORY_DOWNLOAD_TAG, BoundedExecutor
+from s3transfer.utils import CallArgs, OSUtils
+from tests import (
+    BaseSubmissionTaskTest,
+    BaseTaskTest,
+    FileCreator,
+    NonSeekableWriter,
+    RecordingExecutor,
+    StreamWithError,
+    mock,
+    unittest,
+)
 
 
 class DownloadException(Exception):
     pass
 
 
-class WriteCollector(object):
+class WriteCollector:
     """A utility to collect information about writes and seeks"""
+
     def __init__(self):
         self._pos = 0
         self.writes = []
 
-    def seek(self, pos):
+    def seek(self, pos, whence=0):
         self._pos = pos
 
     def write(self, data):
@@ -59,15 +68,23 @@ class WriteCollector(object):
         self._pos += len(data)
 
 
-class CancelledStreamWrapper(object):
+class AlwaysIndicatesSpecialFileOSUtils(OSUtils):
+    """OSUtil that always returns True for is_special_file"""
+
+    def is_special_file(self, filename):
+        return True
+
+
+class CancelledStreamWrapper:
     """A wrapper to trigger a cancellation while stream reading
 
     Forces the transfer coordinator to cancel after a certain amount of reads
     :param stream: The underlying stream to read from
     :param transfer_coordinator: The coordinator for the transfer
-    :param num_reads: On which read to sigal a cancellation. 0 is the first
+    :param num_reads: On which read to signal a cancellation. 0 is the first
         read.
     """
+
     def __init__(self, stream, transfer_coordinator, num_reads=0):
         self._stream = stream
         self._transfer_coordinator = transfer_coordinator
@@ -83,7 +100,7 @@ class CancelledStreamWrapper(object):
 
 class BaseDownloadOutputManagerTest(BaseTaskTest):
     def setUp(self):
-        super(BaseDownloadOutputManagerTest, self).setUp()
+        super().setUp()
         self.osutil = OSUtils()
 
         # Create a file to write to
@@ -95,27 +112,33 @@ class BaseDownloadOutputManagerTest(BaseTaskTest):
         self.io_executor = BoundedExecutor(1000, 1)
 
     def tearDown(self):
-        super(BaseDownloadOutputManagerTest, self).tearDown()
+        super().tearDown()
         shutil.rmtree(self.tempdir)
 
 
 class TestDownloadFilenameOutputManager(BaseDownloadOutputManagerTest):
     def setUp(self):
-        super(TestDownloadFilenameOutputManager, self).setUp()
+        super().setUp()
         self.download_output_manager = DownloadFilenameOutputManager(
-            self.osutil, self.transfer_coordinator,
-            io_executor=self.io_executor)
+            self.osutil,
+            self.transfer_coordinator,
+            io_executor=self.io_executor,
+        )
 
     def test_is_compatible(self):
         self.assertTrue(
-            self.download_output_manager.is_compatible(self.filename))
+            self.download_output_manager.is_compatible(
+                self.filename, self.osutil
+            )
+        )
 
     def test_get_download_task_tag(self):
         self.assertIsNone(self.download_output_manager.get_download_task_tag())
 
     def test_get_fileobj_for_io_writes(self):
         with self.download_output_manager.get_fileobj_for_io_writes(
-                self.future) as f:
+            self.future
+        ) as f:
             # Ensure it is a file like object returned
             self.assertTrue(hasattr(f, 'read'))
             self.assertTrue(hasattr(f, 'seek'))
@@ -126,7 +149,8 @@ class TestDownloadFilenameOutputManager(BaseDownloadOutputManagerTest):
     def test_get_final_io_task(self):
         ref_contents = b'my_contents'
         with self.download_output_manager.get_fileobj_for_io_writes(
-                self.future) as f:
+            self.future
+        ) as f:
             temp_filename = f.name
             # Write some data to test that the data gets moved over to the
             # final location.
@@ -145,19 +169,84 @@ class TestDownloadFilenameOutputManager(BaseDownloadOutputManagerTest):
     def test_can_queue_file_io_task(self):
         fileobj = WriteCollector()
         self.download_output_manager.queue_file_io_task(
-            fileobj=fileobj, data='foo', offset=0)
+            fileobj=fileobj, data='foo', offset=0
+        )
         self.download_output_manager.queue_file_io_task(
-            fileobj=fileobj, data='bar', offset=3)
+            fileobj=fileobj, data='bar', offset=3
+        )
+        self.io_executor.shutdown()
+        self.assertEqual(fileobj.writes, [(0, 'foo'), (3, 'bar')])
+
+    def test_get_file_io_write_task(self):
+        fileobj = WriteCollector()
+        io_write_task = self.download_output_manager.get_io_write_task(
+            fileobj=fileobj, data='foo', offset=3
+        )
+        self.assertIsInstance(io_write_task, IOWriteTask)
+
+        io_write_task()
+        self.assertEqual(fileobj.writes, [(3, 'foo')])
+
+
+class TestDownloadSpecialFilenameOutputManager(BaseDownloadOutputManagerTest):
+    def setUp(self):
+        super().setUp()
+        self.osutil = AlwaysIndicatesSpecialFileOSUtils()
+        self.download_output_manager = DownloadSpecialFilenameOutputManager(
+            self.osutil,
+            self.transfer_coordinator,
+            io_executor=self.io_executor,
+        )
+
+    def test_is_compatible_for_special_file(self):
+        self.assertTrue(
+            self.download_output_manager.is_compatible(
+                self.filename, AlwaysIndicatesSpecialFileOSUtils()
+            )
+        )
+
+    def test_is_not_compatible_for_non_special_file(self):
+        self.assertFalse(
+            self.download_output_manager.is_compatible(
+                self.filename, OSUtils()
+            )
+        )
+
+    def test_get_fileobj_for_io_writes(self):
+        with self.download_output_manager.get_fileobj_for_io_writes(
+            self.future
+        ) as f:
+            # Ensure it is a file like object returned
+            self.assertTrue(hasattr(f, 'read'))
+            # Make sure the name of the file returned is the same as the
+            # final filename as we should not be writing to a temporary file.
+            self.assertEqual(f.name, self.filename)
+
+    def test_get_final_io_task(self):
+        self.assertIsInstance(
+            self.download_output_manager.get_final_io_task(), IOCloseTask
+        )
+
+    def test_can_queue_file_io_task(self):
+        fileobj = WriteCollector()
+        self.download_output_manager.queue_file_io_task(
+            fileobj=fileobj, data='foo', offset=0
+        )
+        self.download_output_manager.queue_file_io_task(
+            fileobj=fileobj, data='bar', offset=3
+        )
         self.io_executor.shutdown()
         self.assertEqual(fileobj.writes, [(0, 'foo'), (3, 'bar')])
 
 
 class TestDownloadSeekableOutputManager(BaseDownloadOutputManagerTest):
     def setUp(self):
-        super(TestDownloadSeekableOutputManager, self).setUp()
+        super().setUp()
         self.download_output_manager = DownloadSeekableOutputManager(
-            self.osutil, self.transfer_coordinator,
-            io_executor=self.io_executor)
+            self.osutil,
+            self.transfer_coordinator,
+            io_executor=self.io_executor,
+        )
 
         # Create a fileobj to write to
         self.fileobj = open(self.filename, 'wb')
@@ -167,18 +256,24 @@ class TestDownloadSeekableOutputManager(BaseDownloadOutputManagerTest):
 
     def tearDown(self):
         self.fileobj.close()
-        super(TestDownloadSeekableOutputManager, self).tearDown()
+        super().tearDown()
 
     def test_is_compatible(self):
         self.assertTrue(
-            self.download_output_manager.is_compatible(self.fileobj))
+            self.download_output_manager.is_compatible(
+                self.fileobj, self.osutil
+            )
+        )
 
     def test_is_compatible_bytes_io(self):
         self.assertTrue(
-            self.download_output_manager.is_compatible(six.BytesIO()))
+            self.download_output_manager.is_compatible(BytesIO(), self.osutil)
+        )
 
     def test_not_compatible_for_non_filelike_obj(self):
-        self.assertFalse(self.download_output_manager.is_compatible(object()))
+        self.assertFalse(
+            self.download_output_manager.is_compatible(object(), self.osutil)
+        )
 
     def test_get_download_task_tag(self):
         self.assertIsNone(self.download_output_manager.get_download_task_tag())
@@ -186,59 +281,82 @@ class TestDownloadSeekableOutputManager(BaseDownloadOutputManagerTest):
     def test_get_fileobj_for_io_writes(self):
         self.assertIs(
             self.download_output_manager.get_fileobj_for_io_writes(
-                self.future),
-            self.fileobj
+                self.future
+            ),
+            self.fileobj,
         )
 
     def test_get_final_io_task(self):
         self.assertIsInstance(
             self.download_output_manager.get_final_io_task(),
-            CompleteDownloadNOOPTask
+            CompleteDownloadNOOPTask,
         )
 
     def test_can_queue_file_io_task(self):
         fileobj = WriteCollector()
         self.download_output_manager.queue_file_io_task(
-            fileobj=fileobj, data='foo', offset=0)
+            fileobj=fileobj, data='foo', offset=0
+        )
         self.download_output_manager.queue_file_io_task(
-            fileobj=fileobj, data='bar', offset=3)
+            fileobj=fileobj, data='bar', offset=3
+        )
         self.io_executor.shutdown()
         self.assertEqual(fileobj.writes, [(0, 'foo'), (3, 'bar')])
 
+    def test_get_file_io_write_task(self):
+        fileobj = WriteCollector()
+        io_write_task = self.download_output_manager.get_io_write_task(
+            fileobj=fileobj, data='foo', offset=3
+        )
+        self.assertIsInstance(io_write_task, IOWriteTask)
+
+        io_write_task()
+        self.assertEqual(fileobj.writes, [(3, 'foo')])
+
 
 class TestDownloadNonSeekableOutputManager(BaseDownloadOutputManagerTest):
     def setUp(self):
-        super(TestDownloadNonSeekableOutputManager, self).setUp()
+        super().setUp()
         self.download_output_manager = DownloadNonSeekableOutputManager(
-            self.osutil, self.transfer_coordinator, io_executor=None)
+            self.osutil, self.transfer_coordinator, io_executor=None
+        )
 
     def test_is_compatible_with_seekable_stream(self):
         with open(self.filename, 'wb') as f:
-            self.assertTrue(self.download_output_manager.is_compatible(f))
+            self.assertTrue(
+                self.download_output_manager.is_compatible(f, self.osutil)
+            )
 
     def test_not_compatible_with_filename(self):
-        self.assertFalse(self.download_output_manager.is_compatible(
-            self.filename))
+        self.assertFalse(
+            self.download_output_manager.is_compatible(
+                self.filename, self.osutil
+            )
+        )
 
     def test_compatible_with_non_seekable_stream(self):
-        class NonSeekable(object):
+        class NonSeekable:
             def write(self, data):
                 pass
 
         f = NonSeekable()
-        self.assertTrue(self.download_output_manager.is_compatible(f))
+        self.assertTrue(
+            self.download_output_manager.is_compatible(f, self.osutil)
+        )
 
     def test_is_compatible_with_bytesio(self):
         self.assertTrue(
-            self.download_output_manager.is_compatible(six.BytesIO()))
+            self.download_output_manager.is_compatible(BytesIO(), self.osutil)
+        )
 
     def test_get_download_task_tag(self):
         self.assertIs(
             self.download_output_manager.get_download_task_tag(),
-            IN_MEMORY_DOWNLOAD_TAG)
+            IN_MEMORY_DOWNLOAD_TAG,
+        )
 
     def test_submit_writes_from_internal_queue(self):
-        class FakeQueue(object):
+        class FakeQueue:
             def request_writes(self, offset, data):
                 return [
                     {'offset': 0, 'data': 'foo'},
@@ -248,18 +366,30 @@ class TestDownloadNonSeekableOutputManager(BaseDownloadOutputManagerTest):
         q = FakeQueue()
         io_executor = BoundedExecutor(1000, 1)
         manager = DownloadNonSeekableOutputManager(
-            self.osutil, self.transfer_coordinator, io_executor=io_executor,
-            defer_queue=q)
+            self.osutil,
+            self.transfer_coordinator,
+            io_executor=io_executor,
+            defer_queue=q,
+        )
         fileobj = WriteCollector()
-        manager.queue_file_io_task(
-            fileobj=fileobj, data='foo', offset=1)
+        manager.queue_file_io_task(fileobj=fileobj, data='foo', offset=1)
         io_executor.shutdown()
         self.assertEqual(fileobj.writes, [(0, 'foo'), (3, 'bar')])
 
+    def test_get_file_io_write_task(self):
+        fileobj = WriteCollector()
+        io_write_task = self.download_output_manager.get_io_write_task(
+            fileobj=fileobj, data='foo', offset=1
+        )
+        self.assertIsInstance(io_write_task, IOStreamingWriteTask)
+
+        io_write_task()
+        self.assertEqual(fileobj.writes, [(0, 'foo')])
+
 
 class TestDownloadSubmissionTask(BaseSubmissionTaskTest):
     def setUp(self):
-        super(TestDownloadSubmissionTask, self).setUp()
+        super().setUp()
         self.tempdir = tempfile.mkdtemp()
         self.filename = os.path.join(self.tempdir, 'myfile')
 
@@ -270,7 +400,7 @@ class TestDownloadSubmissionTask(BaseSubmissionTaskTest):
 
         # Create a stream to read from
         self.content = b'my content'
-        self.stream = six.BytesIO(self.content)
+        self.stream = BytesIO(self.content)
 
         # A list to keep track of all of the bodies sent over the wire
         # and their order.
@@ -284,19 +414,21 @@ class TestDownloadSubmissionTask(BaseSubmissionTaskTest):
             'osutil': self.osutil,
             'request_executor': self.executor,
             'io_executor': self.io_executor,
-            'transfer_future': self.transfer_future
+            'transfer_future': self.transfer_future,
         }
         self.submission_task = self.get_download_submission_task()
 
     def tearDown(self):
-        super(TestDownloadSubmissionTask, self).tearDown()
+        super().tearDown()
         shutil.rmtree(self.tempdir)
 
     def get_call_args(self, **kwargs):
         default_call_args = {
-            'fileobj': self.filename, 'bucket': self.bucket,
-            'key': self.key, 'extra_args': self.extra_args,
-            'subscribers': self.subscribers
+            'fileobj': self.filename,
+            'bucket': self.bucket,
+            'key': self.key,
+            'extra_args': self.extra_args,
+            'subscribers': self.subscribers,
         }
         default_call_args.update(kwargs)
         return CallArgs(**default_call_args)
@@ -316,21 +448,21 @@ class TestDownloadSubmissionTask(BaseSubmissionTaskTest):
             # If it was ranged get, make sure we do not include the join task.
             submissions_to_compare = submissions_to_compare[:-1]
         for submission in submissions_to_compare:
-            self.assertEqual(
-                submission['tag'], tag_value)
+            self.assertEqual(submission['tag'], tag_value)
 
     def add_head_object_response(self):
         self.stubber.add_response(
-            'head_object', {'ContentLength': len(self.content)})
+            'head_object', {'ContentLength': len(self.content)}
+        )
 
     def add_get_responses(self):
         chunksize = self.config.multipart_chunksize
         for i in range(0, len(self.content), chunksize):
             if i + chunksize > len(self.content):
-                stream = six.BytesIO(self.content[i:])
+                stream = BytesIO(self.content[i:])
                 self.stubber.add_response('get_object', {'Body': stream})
             else:
-                stream = six.BytesIO(self.content[i:i+chunksize])
+                stream = BytesIO(self.content[i : i + chunksize])
                 self.stubber.add_response('get_object', {'Body': stream})
 
     def configure_for_ranged_get(self):
@@ -339,7 +471,8 @@ class TestDownloadSubmissionTask(BaseSubmissionTaskTest):
 
     def get_download_submission_task(self):
         return self.get_task(
-            DownloadSubmissionTask, main_kwargs=self.submission_main_kwargs)
+            DownloadSubmissionTask, main_kwargs=self.submission_main_kwargs
+        )
 
     def wait_and_assert_completed_successfully(self, submission_task):
         submission_task()
@@ -432,7 +565,7 @@ class TestDownloadSubmissionTask(BaseSubmissionTaskTest):
 
 class TestGetObjectTask(BaseTaskTest):
     def setUp(self):
-        super(TestGetObjectTask, self).setUp()
+        super().setUp()
         self.bucket = 'mybucket'
         self.key = 'mykey'
         self.extra_args = {}
@@ -440,24 +573,30 @@ class TestGetObjectTask(BaseTaskTest):
         self.max_attempts = 5
         self.io_executor = BoundedExecutor(1000, 1)
         self.content = b'my content'
-        self.stream = six.BytesIO(self.content)
+        self.stream = BytesIO(self.content)
         self.fileobj = WriteCollector()
         self.osutil = OSUtils()
-        self.io_chunksize = 64 * (1024 ** 2)
+        self.io_chunksize = 64 * (1024**2)
+        self.task_cls = GetObjectTask
         self.download_output_manager = DownloadSeekableOutputManager(
-            self.osutil, self.transfer_coordinator, self.io_executor)
+            self.osutil, self.transfer_coordinator, self.io_executor
+        )
 
     def get_download_task(self, **kwargs):
         default_kwargs = {
-            'client': self.client, 'bucket': self.bucket, 'key': self.key,
-            'fileobj': self.fileobj, 'extra_args': self.extra_args,
+            'client': self.client,
+            'bucket': self.bucket,
+            'key': self.key,
+            'fileobj': self.fileobj,
+            'extra_args': self.extra_args,
             'callbacks': self.callbacks,
             'max_attempts': self.max_attempts,
             'download_output_manager': self.download_output_manager,
             'io_chunksize': self.io_chunksize,
         }
         default_kwargs.update(kwargs)
-        return self.get_task(GetObjectTask, main_kwargs=default_kwargs)
+        self.transfer_coordinator.set_status_to_queued()
+        return self.get_task(self.task_cls, main_kwargs=default_kwargs)
 
     def assert_io_writes(self, expected_writes):
         # Let the io executor process all of the writes before checking
@@ -467,8 +606,9 @@ class TestGetObjectTask(BaseTaskTest):
 
     def test_main(self):
         self.stubber.add_response(
-            'get_object', service_response={'Body': self.stream},
-            expected_params={'Bucket': self.bucket, 'Key': self.key}
+            'get_object',
+            service_response={'Body': self.stream},
+            expected_params={'Bucket': self.bucket, 'Key': self.key},
         )
         task = self.get_download_task()
         task()
@@ -478,10 +618,13 @@ class TestGetObjectTask(BaseTaskTest):
 
     def test_extra_args(self):
         self.stubber.add_response(
-            'get_object', service_response={'Body': self.stream},
+            'get_object',
+            service_response={'Body': self.stream},
             expected_params={
-                'Bucket': self.bucket, 'Key': self.key, 'Range': 'bytes=0-'
-            }
+                'Bucket': self.bucket,
+                'Key': self.key,
+                'Range': 'bytes=0-',
+            },
         )
         self.extra_args['Range'] = 'bytes=0-'
         task = self.get_download_task()
@@ -492,8 +635,9 @@ class TestGetObjectTask(BaseTaskTest):
 
     def test_control_chunk_size(self):
         self.stubber.add_response(
-            'get_object', service_response={'Body': self.stream},
-            expected_params={'Bucket': self.bucket, 'Key': self.key}
+            'get_object',
+            service_response={'Body': self.stream},
+            expected_params={'Bucket': self.bucket, 'Key': self.key},
         )
         task = self.get_download_task(io_chunksize=1)
         task()
@@ -501,14 +645,15 @@ class TestGetObjectTask(BaseTaskTest):
         self.stubber.assert_no_pending_responses()
         expected_contents = []
         for i in range(len(self.content)):
-            expected_contents.append((i, bytes(self.content[i:i+1])))
+            expected_contents.append((i, bytes(self.content[i : i + 1])))
 
         self.assert_io_writes(expected_contents)
 
     def test_start_index(self):
         self.stubber.add_response(
-            'get_object', service_response={'Body': self.stream},
-            expected_params={'Bucket': self.bucket, 'Key': self.key}
+            'get_object',
+            service_response={'Body': self.stream},
+            expected_params={'Bucket': self.bucket, 'Key': self.key},
         )
         task = self.get_download_task(start_index=5)
         task()
@@ -516,16 +661,35 @@ class TestGetObjectTask(BaseTaskTest):
         self.stubber.assert_no_pending_responses()
         self.assert_io_writes([(5, self.content)])
 
+    def test_uses_bandwidth_limiter(self):
+        bandwidth_limiter = mock.Mock(BandwidthLimiter)
+
+        self.stubber.add_response(
+            'get_object',
+            service_response={'Body': self.stream},
+            expected_params={'Bucket': self.bucket, 'Key': self.key},
+        )
+        task = self.get_download_task(bandwidth_limiter=bandwidth_limiter)
+        task()
+
+        self.stubber.assert_no_pending_responses()
+        self.assertEqual(
+            bandwidth_limiter.get_bandwith_limited_stream.call_args_list,
+            [mock.call(mock.ANY, self.transfer_coordinator)],
+        )
+
     def test_retries_succeeds(self):
         self.stubber.add_response(
-            'get_object', service_response={
+            'get_object',
+            service_response={
                 'Body': StreamWithError(self.stream, SOCKET_ERROR)
             },
-            expected_params={'Bucket': self.bucket, 'Key': self.key}
+            expected_params={'Bucket': self.bucket, 'Key': self.key},
         )
         self.stubber.add_response(
-            'get_object', service_response={'Body': self.stream},
-            expected_params={'Bucket': self.bucket, 'Key': self.key}
+            'get_object',
+            service_response={'Body': self.stream},
+            expected_params={'Bucket': self.bucket, 'Key': self.key},
         )
         task = self.get_download_task()
         task()
@@ -538,10 +702,11 @@ class TestGetObjectTask(BaseTaskTest):
     def test_retries_failure(self):
         for _ in range(self.max_attempts):
             self.stubber.add_response(
-                'get_object', service_response={
+                'get_object',
+                service_response={
                     'Body': StreamWithError(self.stream, SOCKET_ERROR)
                 },
-                expected_params={'Bucket': self.bucket, 'Key': self.key}
+                expected_params={'Bucket': self.bucket, 'Key': self.key},
             )
 
         task = self.get_download_task()
@@ -556,22 +721,25 @@ class TestGetObjectTask(BaseTaskTest):
     def test_retries_in_middle_of_streaming(self):
         # After the first read a retryable error will be thrown
         self.stubber.add_response(
-            'get_object', service_response={
+            'get_object',
+            service_response={
                 'Body': StreamWithError(
-                    copy.deepcopy(self.stream), SOCKET_ERROR, 1)
+                    copy.deepcopy(self.stream), SOCKET_ERROR, 1
+                )
             },
-            expected_params={'Bucket': self.bucket, 'Key': self.key}
+            expected_params={'Bucket': self.bucket, 'Key': self.key},
         )
         self.stubber.add_response(
-            'get_object', service_response={'Body': self.stream},
-            expected_params={'Bucket': self.bucket, 'Key': self.key}
+            'get_object',
+            service_response={'Body': self.stream},
+            expected_params={'Bucket': self.bucket, 'Key': self.key},
         )
         task = self.get_download_task(io_chunksize=1)
         task()
 
         self.stubber.assert_no_pending_responses()
         expected_contents = []
-        # This is the content intially read in before the retry hit on the
+        # This is the content initially read in before the retry hit on the
         # second read()
         expected_contents.append((0, bytes(self.content[0:1])))
 
@@ -580,7 +748,7 @@ class TestGetObjectTask(BaseTaskTest):
         # element in the list should be a copy of the first element since
         # a retryable exception happened in between.
         for i in range(len(self.content)):
-            expected_contents.append((i, bytes(self.content[i:i+1])))
+            expected_contents.append((i, bytes(self.content[i : i + 1])))
         self.assert_io_writes(expected_contents)
 
     def test_cancels_out_of_queueing(self):
@@ -588,9 +756,10 @@ class TestGetObjectTask(BaseTaskTest):
             'get_object',
             service_response={
                 'Body': CancelledStreamWrapper(
-                    self.stream, self.transfer_coordinator)
+                    self.stream, self.transfer_coordinator
+                )
             },
-            expected_params={'Bucket': self.bucket, 'Key': self.key}
+            expected_params={'Bucket': self.bucket, 'Key': self.key},
         )
         task = self.get_download_task()
         task()
@@ -601,17 +770,47 @@ class TestGetObjectTask(BaseTaskTest):
         # io queue.
         self.assert_io_writes([])
 
+    def test_handles_callback_on_initial_error(self):
+        # We can't use the stubber for this because we need to raise
+        # a S3_RETRYABLE_DOWNLOAD_ERRORS, and the stubber only allows
+        # you to raise a ClientError.
+        self.client.get_object = mock.Mock(side_effect=SOCKET_ERROR())
+        task = self.get_download_task()
+        task()
+        self.transfer_coordinator.announce_done()
+        # Should have failed out on a RetriesExceededError because
+        # get_object keeps raising a socket error.
+        with self.assertRaises(RetriesExceededError):
+            self.transfer_coordinator.result()
+
+
+class TestImmediatelyWriteIOGetObjectTask(TestGetObjectTask):
+    def setUp(self):
+        super().setUp()
+        self.task_cls = ImmediatelyWriteIOGetObjectTask
+        # When data is written out, it should not use the io executor at all
+        # if it does use the io executor that is a deviation from expected
+        # behavior as the data should be written immediately to the file
+        # object once downloaded.
+        self.io_executor = None
+        self.download_output_manager = DownloadSeekableOutputManager(
+            self.osutil, self.transfer_coordinator, self.io_executor
+        )
+
+    def assert_io_writes(self, expected_writes):
+        self.assertEqual(self.fileobj.writes, expected_writes)
+
 
 class BaseIOTaskTest(BaseTaskTest):
     def setUp(self):
-        super(BaseIOTaskTest, self).setUp()
+        super().setUp()
         self.files = FileCreator()
         self.osutil = OSUtils()
         self.temp_filename = os.path.join(self.files.rootdir, 'mytempfile')
         self.final_filename = os.path.join(self.files.rootdir, 'myfile')
 
     def tearDown(self):
-        super(BaseIOTaskTest, self).tearDown()
+        super().tearDown()
         self.files.remove_all()
 
 
@@ -620,18 +819,12 @@ class TestIOStreamingWriteTask(BaseIOTaskTest):
         with open(self.temp_filename, 'wb') as f:
             task = self.get_task(
                 IOStreamingWriteTask,
-                main_kwargs={
-                    'fileobj': f,
-                    'data': b'foobar'
-                }
+                main_kwargs={'fileobj': f, 'data': b'foobar'},
             )
             task()
             task2 = self.get_task(
                 IOStreamingWriteTask,
-                main_kwargs={
-                    'fileobj': f,
-                    'data': b'baz'
-                }
+                main_kwargs={'fileobj': f, 'data': b'baz'},
             )
             task2()
         with open(self.temp_filename, 'rb') as f:
@@ -646,22 +839,14 @@ class TestIOWriteTask(BaseIOTaskTest):
             # Write once to the file
             task = self.get_task(
                 IOWriteTask,
-                main_kwargs={
-                    'fileobj': f,
-                    'data': b'foo',
-                    'offset': 0
-                }
+                main_kwargs={'fileobj': f, 'data': b'foo', 'offset': 0},
             )
             task()
 
             # Write again to the file
             task = self.get_task(
                 IOWriteTask,
-                main_kwargs={
-                    'fileobj': f,
-                    'data': b'bar',
-                    'offset': 3
-                }
+                main_kwargs={'fileobj': f, 'data': b'bar', 'offset': 3},
             )
             task()
 
@@ -677,14 +862,47 @@ class TestIORenameFileTask(BaseIOTaskTest):
                 main_kwargs={
                     'fileobj': f,
                     'final_filename': self.final_filename,
-                    'osutil': self.osutil
-                }
+                    'osutil': self.osutil,
+                },
             )
             task()
         self.assertTrue(os.path.exists(self.final_filename))
         self.assertFalse(os.path.exists(self.temp_filename))
 
 
+class TestIOCloseTask(BaseIOTaskTest):
+    def test_main(self):
+        with open(self.temp_filename, 'w') as f:
+            task = self.get_task(IOCloseTask, main_kwargs={'fileobj': f})
+            task()
+            self.assertTrue(f.closed)
+
+
+class TestDownloadChunkIterator(unittest.TestCase):
+    def test_iter(self):
+        content = b'my content'
+        body = BytesIO(content)
+        ref_chunks = []
+        for chunk in DownloadChunkIterator(body, len(content)):
+            ref_chunks.append(chunk)
+        self.assertEqual(ref_chunks, [b'my content'])
+
+    def test_iter_chunksize(self):
+        content = b'1234'
+        body = BytesIO(content)
+        ref_chunks = []
+        for chunk in DownloadChunkIterator(body, 3):
+            ref_chunks.append(chunk)
+        self.assertEqual(ref_chunks, [b'123', b'4'])
+
+    def test_empty_content(self):
+        body = BytesIO(b'')
+        ref_chunks = []
+        for chunk in DownloadChunkIterator(body, 3):
+            ref_chunks.append(chunk)
+        self.assertEqual(ref_chunks, [b''])
+
+
 class TestDeferQueue(unittest.TestCase):
     def setUp(self):
         self.q = DeferQueue()
@@ -707,8 +925,8 @@ class TestDeferQueue(unittest.TestCase):
                 {'offset': 0, 'data': 'a'},
                 {'offset': 1, 'data': 'b'},
                 {'offset': 2, 'data': 'c'},
-                {'offset': 3, 'data': 'd'}
-            ]
+                {'offset': 3, 'data': 'd'},
+            ],
         )
 
     def test_unlocks_partial_range(self):
@@ -722,7 +940,7 @@ class TestDeferQueue(unittest.TestCase):
             [
                 {'offset': 0, 'data': 'a'},
                 {'offset': 1, 'data': 'b'},
-            ]
+            ],
         )
 
     def test_data_can_be_any_size(self):
@@ -733,7 +951,7 @@ class TestDeferQueue(unittest.TestCase):
             [
                 {'offset': 0, 'data': 'abcde'},
                 {'offset': 5, 'data': 'hello world'},
-            ]
+            ],
         )
 
     def test_data_queued_in_order(self):
@@ -757,7 +975,7 @@ class TestDeferQueue(unittest.TestCase):
 
         self.assertEqual(
             self.q.request_writes(offset=3, data='d'),
-            [{'offset': 3, 'data': 'd'}]
+            [{'offset': 3, 'data': 'd'}],
         )
 
     def test_duplicate_writes_are_ignored(self):
@@ -777,5 +995,5 @@ class TestDeferQueue(unittest.TestCase):
                 # Note we're seeing 'b' 'c', and not 'X', 'Y'.
                 {'offset': 1, 'data': 'b'},
                 {'offset': 2, 'data': 'c'},
-            ]
+            ],
         )
